import logging as log
import os
from typing import Any

import yaml
from dotenv import load_dotenv
from jinja2 import Template

from . import sources
from .inca import Chasqui
from .utils import ensure_rooting, secrets

class Curie:
    """
    The module we all came here for.
    Curie is a framework for database agnostic information routing.
    Common use cases include:
        - Data synchronization
        - API integration
        - MLOps pipelines
        - Data Lifecycle Management
    
    """
    __version__ = '0.1.0'
    def __init__(self, config_file='./config/pathways.yaml', debug=False, verbose=False, **kwargs):
        config_file = ensure_rooting(config_file)
        # Logging
        loglevel = log.root.level
        if log.root.level == log.NOTSET or log.root.level == log.WARNING:
            if debug:
                loglevel = log.DEBUG
            elif verbose:
                loglevel = log.INFO
        log.basicConfig(
            level=loglevel, format='%(asctime)s - %(levelname)s - %(message)s')
        # Flags
        self.ALL = Chasqui.ALL
        self.name = None

        if os.path.exists(config_file):
            self.__config_file = config_file
            self.__config__ = yaml.safe_load(open(self.__config_file))

            self.__connections = Connections(self.__config__['Connections'])
            self.__resource_tree = SourceTree(
                self.__connections, defer_import=True)

            self.__pathways = Pathways(self.__config__['Pathways'])
        else:
            raise FileNotFoundError(
                'Configuration file not found at ' + config_file)

    def __repr__(self):
        return f'Curie({len(self.__resource_tree)} Connections, {len(self.__pathways)} Pathways)'

    def connect(self, name: str, help=False) -> Any:
        if name not in self.__resource_tree:
            name = name.replace('_', '-')  # Snake case conformity.
        try:
            # Help will print out a snippet to aid in using the provided connection
            return self.__resource_tree[name].connect(help=help)
        except KeyError:
            raise AttributeError(f'No connection named {name}')

    def path(self, name):
        # Stages a pathway for loading
        if name == '.':
            for path in self.__config__['Pathways']:
                self.name = None
                self.__pathways.load_path(path, self)
        elif name not in self.__config__['Pathways']:
            raise AttributeError(f'No pathway named {name}')
        else:
            self.name = name
            self.__pathways.load_path(name, self)
        return self
    
    def show_dag(self, **kwargs):
        if self.name not in self.__pathways:
            raise AttributeError(f'No pathway named {self.name}')
        else:
            print("PATHWAY NAME >>> " + self.name)
            return self.__pathways[self.name].show_dag(**kwargs)

    def execute(self, tables: list):
        """
        Execute one pipeline only, ignore ALL flags
        """
        if self.name is None:
            raise AttributeError(
                'No pathway recognized. Please specify a single pathway using path()')
        if self.name not in self.__pathways:
            raise AttributeError(f'No pathway named {self.name}')
        self.__pathways[self.name].execute(tables)
        return None

    def clean(self, *facet):
        """
        Clean many or one pipelines from a project
        """
        if self.name is None:
            for name in self.__pathways:
                self.__pathways[name].clean(*facet)
        elif self.name not in self.__pathways:
            raise AttributeError(f'No pathway named {self.name}')
        else:
            self.__pathways[self.name].clean(*facet)
        return None

    def mode(self, mode):
        """
        Prepare pipeline using specified mode
        """
        if self.name is None:
            raise AttributeError(
                'No pathway recognized. Please specify a single pathway using path()')
        if self.name not in self.__pathways:
            raise AttributeError(f'No pathway named {self.name}')
        self.__pathways[self.name].mode(mode)
        return self

    @property
    def connections(self):
        return self.__resource_tree


class SourceTree(dict):
    def __init__(self, config, defer_import=True):
        super().__init__()
        self.defer_import = defer_import
        self.prepare(config)

    def prepare(self, config):
        for source in config:
            for conn in config[source]:
                try:
                    temp = {conn: getattr(sources, source)(
                        **config[source][conn], defer_import=self.defer_import)}
                    self.update(temp)
                except Exception as e:
                    log.info("Notice: Could not load source: " + source +
                             " with connection: " + conn + ". Check your configuration file.")
        return self


class Connections(dict):
    def __init__(self, config_dict):
        self.config_dict = config_dict
        self.connection_contents = self.__identify_and_load()

        for key in self.connection_contents:
            for _c in self.connection_contents[key]:
                if 'secrets' in self.connection_contents[key][_c]:
                    secret_manager = list(
                        self.connection_contents[key][_c]['secrets'].keys())[0]
                    args = getattr(secrets, secret_manager)(
                        **self.connection_contents[key][_c]['secrets'][secret_manager])
                    for kkey in self.connection_contents[key][_c]:
                        if kkey != 'secrets':
                            self.connection_contents[key][_c][kkey] = Template(
                                self.connection_contents[key][_c][kkey]).render(**args)
                    del self.connection_contents[key][_c]['secrets']
        super().__init__(self.connection_contents)

    def __load_from_file(self, obj):
        """
        Load a configuration file from a file path
        """
        self.__file = ensure_rooting(obj)
        if os.path.exists(self.__file):
            return yaml.safe_load(open(self.__file))

    def __identify_and_load(self):
        """
        Identify the loader to use for the configuration file and load it into a dictionary
        """
        __valid_handlers = {
            'file': lambda x: self.__load_from_file(x)
        }
        for key in __valid_handlers:
            if key in self.config_dict:
                return __valid_handlers[key](self.config_dict[key])
        raise ValueError('No valid configuration file handler found')


class Blueprint(Chasqui):
    def __init__(self, connection, blueprint, curie=None, **kwargs):
        """
        An abstraction of Chasqui
        Overlays utility functions on the Chasqui class to make it easier to use, and harder to break

        Args:
            connection (str): Connection name
            blueprint (str): Path to blueprint file
            curie (Curie, optional): Curie object. Defaults to None.

        Raises:
            AttributeError: Blueprint file not found

        """
        if curie is None:
            raise AttributeError('Curie object not passed to Blueprint')
        self.connection = connection
        if os.path.exists(ensure_rooting(blueprint)):
            self.blueprint_file = ensure_rooting(blueprint)
            self.blueprint = yaml.safe_load(open(self.blueprint_file))
            super().__init__(self.blueprint)
            if 'name' in kwargs:
                self.name = kwargs['name']
            self.source = curie.connections[self.connection]
        else:
            raise FileNotFoundError('Blueprint file not found at ' + blueprint)

    def mode(self, mode='save'):
        """
        Prepare pipeline using specified mode

        Args:
            mode (str, optional): Mode to prepare pipeline for. Defaults to 'save'.
        """
        self.mode_ = mode
        super().prepare(mode=mode)
        return self

    def execute(self, tables: list):
        """
        Execute the pipeline

        Args:
            tables (list): List of tables to execute
        """
        super().execute(tables)
        return self
    
    def show_dag(self, **kwargs):
        return super().show_dag(**kwargs)

    def clean(self, facet):
        super().clean(facet)


class Pathways(dict):
    def __init__(self, config_dict):
        """
        * Pathways is a dictionary of Blueprint objects
        * They are loaded from a configuration file and accessed by name
        """
        super().__init__()
        self.config_dict = config_dict

    def load_path(self, __name, __curie):
        """
        * Load a blueprint by name and return a Blueprint object
        """
        super().update({__name: Blueprint(curie=__curie,
                                          **self.config_dict[__name], name=__name)})
        return self[__name]
    
    def show_dag(self,__name, **kwargs):
        print("PATHWAY NAME >>> " + __name)
        return self[__name].show_dag(**kwargs)

    def __str__(self):
        return super().__str__()



import os

def find_root(landmark: str = 'project.yaml') -> str:
    """
    Find the root directory of a project.
    """
    root = os.getcwd()
    while True:
        if os.path.exists(os.path.join(root, landmark)):
            return root
        else:
            root = os.path.dirname(root)
            # Test if we have reached the root of the drive
            if root == os.path.dirname(root):
                raise FileNotFoundError(f'Could not identify the project root. Please ensure that {landmark} exists in the project root.')
    return None

__project_root__ = find_root()

def ensure_rooting(path) -> str:
    """
    Ensure that the path is rooted to the project.
    """
    if os.path.isabs(path) and path.startswith(__project_root__):
        return path
    else:
        return os.path.normpath(os.path.join(__project_root__, path))



import importlib.resources as pkg_resources
import inspect
import logging as log
import os
from functools import wraps
from typing import Any

import yaml
from dotenv import load_dotenv
from jinja2 import Template
import jinja2

from .utils import ensure_rooting, secrets

global loader, jenv
loader = jinja2.FileSystemLoader(searchpath="./")
jenv = jinja2.Environment(loader=loader)

def import_raw(path):
    return open(ensure_rooting(path)).read()

def title(s):
    return s.title()

jenv.globals.update(
    import_raw=import_raw,
    title=title
)

class Documentation:
    def __init__(self, config="./config/pathways.yaml"):
        
        self.config = ensure_rooting(config)
        self.pathways = yaml.safe_load(open(self.config).read())

        self.site_config = self.pathways['Documentation']
        self.site_config['site']['nav'] = {'pipelines': []}

        self.replace_index()

        pipes = list( self.render_pathways(self.pathways)['Pathways'].keys() )

        for pipe in pipes:
            blueprint_file = ensure_rooting(self.pathways['Pathways'][pipe]['blueprint'])
            if not os.path.exists(blueprint_file):
                raise FileNotFoundError(f"Blueprint file for {pipe} not found at {blueprint_file}.")
            blueprint = yaml.safe_load(open(blueprint_file).read())
            self.render_pipe(pipe, blueprint)
        
        self.render_site_config()
        print(self.site_config)
    
    def document(template, endpoint):
        def decorator(func):
            def wrapper(*args, **kwargs):
                mapped_args = dict(zip(inspect.getfullargspec(func).args, args))
                mapped_args.update(kwargs)
                _template = jenv.from_string(pkg_resources.read_text('curie.defaults.jinja', template))
                _endpoint = ensure_rooting(endpoint.format(**mapped_args))
                results = func(*args, **kwargs)
                if not os.path.exists(os.path.dirname(_endpoint)):
                    os.makedirs(os.path.dirname(_endpoint))
                with open(_endpoint, 'w') as f:
                    f.write(_template.render(**results))
                return results
            return wrapper
        return decorator

    def replace_index(self, readme='./README.md', endpoint="./docs/index.md"):
        src = ensure_rooting(readme)
        dst = ensure_rooting(endpoint)
        with open(dst, 'w') as f:
            f.write(open(src).read())
    
    @document('pipeline.md.j2', './docs/pipelines.md')
    def render_pathways(self, pathways):
        return pathways
    
    @document('pipeline-bp.md.j2', './docs/pipes/{name}.md')
    def render_pipe(self, name, pipeline_def):
        pipeline_def['pipeline_name'] = name
        self.site_config['site']['nav']['pipelines'].append({'name': name, 'link': f'pipes/{name}.md'})
        return pipeline_def
    
    @document('docs-settings.yaml.j2', 'docs-settings.yaml')
    def render_site_config(self):
        return self.site_config


import argparse
import importlib.resources as pkg_resources
import logging as log
import os
import subprocess
import sys
import time
from zipfile import ZipFile
import yaml

try:
    from .lab import Curie
    from .mk import Documentation
    from .utils import ensure_rooting
except FileNotFoundError as e:
    pass # this is fine, other errors will be raised if there is a problem. Doing this allows us to initialize a project without the lab and mk modules.

def generate_docs(args):
    try:
        subprocess.run(["mkdocs", "build", "--config-file", "docs-settings.yaml"])
        log.info("Documentation generated using mkdocs.")
        docs = Documentation()
    except Exception as e:
        log.error("Error generating documentation:", str(e))

def serve_docs(args):
    try:
        docs = Documentation()
        subprocess.run(["mkdocs", "serve", "--config-file", "docs-settings.yaml"])
    except Exception as e:
        log.error("Error serving documentation:", str(e))

def __initialize_project(args):
    # initialize a new project
        print("Initializing new project: " + args.init)
        # check if project already exists
        if os.path.exists(args.init):
            return log.error("Error: Project already exists.")
        else:
            os.mkdir(args.init)
            __defaults__ = pkg_resources.files("curie.defaults").joinpath("curie_seed.zip")
            with ZipFile(__defaults__, "r") as zipFile:
                zipFile.extractall(path=args.init)
            log.info("Success: Project initialized.")

def add_pipeline(args):
    project_def_file = ensure_rooting('./project.yaml')
    project_def = yaml.safe_load(open(project_def_file).read())['Configuration']
    connections_file, pathways_file = [ensure_rooting(project_def[x]) for x in ['connections', 'pathways']]

    print("Adding pipeline: " + args.name)
    # if a connection is specified, check if it exists
    connection_found = False
    if hasattr(args, "connection") and args.connection is not None:
        connections = yaml.safe_load(open(connections_file).read())
        for dbtype in connections:
            if args.connection in connections[dbtype]:
                connection_found = True
                break
        if not connection_found:
            return log.error("Error: Connection not found.")
        
    # check if pipeline already exists
    if os.path.exists(pathways_file):
        pathways = yaml.safe_load(open(pathways_file).read())
        name_in_files = lambda n,f: any([n in x for x in f])
        if args.name in pathways['Pathways'] or name_in_files(args.name, os.listdir('./blueprints')):
            return log.error("Error: Pipeline already exists.")
        else:
            pathways['Pathways'][args.name] = {
                'connection': '' if not hasattr(args, "connection") or args.connection is None else args.connection,
                'blueprint': f'./blueprints/{args.name}.yaml',
                'description': ''
            }
            with open(pathways_file, 'w') as f:
                yaml.dump(pathways, f)

            # create blueprint
            blueprint_file = ensure_rooting(f'./blueprints/{args.name}.yaml')
            with open(blueprint_file, 'w') as f:
                yaml.dump({
                    "sync": {
                        "target": "default",
                        "overwrite": True
                    },
                    "arguments": {'limit':100},
                    "etl": {
                        "some_table":{
                            "manifest": 'table',
                            "schema": "public",
                            "save": {
                                "query": "SELECT * FROM {{this}}"
                            }
                        }
                    }
                }, f)
    pass


def main():
    # initialize parser
    parser = argparse.ArgumentParser(description="Curie - A Python-based framework for database agnostic information routing.", prog="curie")
    subparsers = parser.add_subparsers(help="Curie commands", dest="command", required=True)

    # docs arguments
    docs_parser = subparsers.add_parser("docs", help="Documentation commands")
    docs_subparsers = docs_parser.add_subparsers(help="Documentation commands", dest="docs_command")

    docs_parser_generate = docs_subparsers.add_parser("generate", help="Generate documentation")
    docs_parser_generate.set_defaults(func=generate_docs)

    docs_parser_serve = docs_subparsers.add_parser("serve", help="Serve documentation")
    docs_parser_serve.set_defaults(func=serve_docs)

    # etl arguments
    etl_parser = subparsers.add_parser("etl", help="ETL commands")
    etl_subparsers = etl_parser.add_subparsers(help="ETL commands", dest="etl_command")
    
    # ETL functions

    # Add a pipeline
    etl_parser_add = etl_subparsers.add_parser("add", help="Add a pipeline")
    etl_parser_add.add_argument('-n', '--name', help="Name of the pipeline", action="store", metavar="<pipeline_name>")
    etl_parser_add.add_argument('-c', '--connection', help="Connection to use for the pipeline", action="store", metavar="<connection_name>")
    etl_parser_add.set_defaults(func=add_pipeline)

    # ETL arguments
    etl_parser.add_argument('-v','--version', help="Print the version of Curie", action="store_true")
    etl_parser.add_argument('-i', '--init', help="Initialize a new project", action="store", metavar="<project_name>")

    etl_parser.add_argument('-r', '--run', help="Run the pipeline against the database", action="store", metavar="<pipeline_name>")
    etl_parser.add_argument('-s', '--save', help="Save tables locally", action="store", metavar="<pipeline_name>")

    etl_parser.add_argument('--clean', help="Clean a project", action="store", metavar="<pipeline_name>")
    etl_parser.add_argument('-c','--compile', help="Compile a pipeline", action="store_true", default=False)
    
    etl_parser.add_argument('-t','--tables', nargs="+", help="List of tables to affect. For all use \".\"", action="store", metavar="<table>")
    etl_parser.add_argument('--facet', nargs="+", help="List of facets to clean (compiled, data). For all use \".\"", action="store", metavar="<facet>") # Not required

    etl_parser.add_argument('--debug', help="Enable debug mode", action="store_const", dest="loglevel", const=log.DEBUG, default=log.WARNING)
    etl_parser.add_argument('--verbose', help="Enable verbose mode", action="store_const", dest="loglevel", const=log.INFO)
    etl_parser.add_argument('--show-dag', help="Generate a DAG of the pipeline", action="store_true", default=False)

    def dagman(args,prep):
        if args.show_dag:
            log.info("DAG ++ " + str(prep.show_dag(use_filenames=True)))

    # USE SUBPARSERS
    args, extras = parser.parse_known_args()


    if hasattr(args, "func"):
        print("Running command: " + args.command)
        args.func(args)
    elif args.command == "etl":
        log.basicConfig(level=args.loglevel, format='%(asctime)s - %(levelname)s - %(message)s')
        args.tables = args.tables if args.tables is not None else ['.']
        if args.init:
            __initialize_project(args)
            return
        elif args.clean:
            # curie --clean <pipeline_name> --facet <facet(s)>
            # Defaults to all facets: curie --clean <pipeline_name>
            pipeline = args.clean
            facets = ['.']
            if args.facet is not None:
                facets = args.facet
            log.info("Cleaning project: " + pipeline)
            curie_ = Curie()
            curie_.path(pipeline).clean(facets)
            return
        elif args.version:
            # curie --version
            print("Curie version: " + Curie.__version__)
            return
        elif args.compile:
            curie_ = Curie()
            if args.run:
                # curie --run <pipeline_name> --compile
                # curie -r <pipeline_name> -c
                com = curie_.path(args.run).mode('run')
                dagman(args,com)
            elif args.save:
                # curie --save <pipeline_name> --compile
                # curie -s <pipeline_name> -c
                com = curie_.path(args.save).mode('save')
                dagman(args,com)
            return
        elif args.run:
            # curie --run <pipeline_name> -t <table(s)>
            # curie -r <pipeline_name> -t <table(s)>
            curie_ = Curie()
            curie_.path(args.run).mode('run').execute(args.tables)
            return
        elif args.save:
            # curie -s <pipeline_name> -t <table(s)>
            curie_ = Curie()
            curie_.path(args.save).mode('save').execute(args.tables)
            return
    else:
        parser.print_help()

if __name__ == "__main__":
    main()



import logging as log
import os
import shutil

from jinja2 import Template
from .utils import ensure_rooting


class Chasqui:
    ALL = '.'
    def __init__(self, topology, **kwargs):
        """
        Args:
            topology (dict): A dictionary containing the ETL topology. AKA the config.
            kwargs (dict): A dictionary containing the arguments for the ETL topology.
        """
        self.topology = topology
        self.ARGUMENTS = topology['arguments']
        self.preprocessed = None
        self.kwargs = kwargs
        self._name = "unnamed"
        self._source = None
        self._mode = None

    def prepare(self, mode=None):
        """
        Prepare the ETL topology for execution.

        Args:
            mode (str): The mode to prepare the ETL topology for. (run, save)
        """
        if mode is None:
            raise Exception("Error: No mode specified. Please specify a mode.")
        self._mode = mode
        self.dag = self.__infer_dag(self.topology['etl'], mode=mode)
        self.dag_roots = self.__dag_roots(self.topology['etl'], mode=mode)
        self.preprocessed = self.__prepare_etl(self.topology['etl'],self.dag, mode=mode, **self.ARGUMENTS)
        self.__handle_compiled_queries(self.preprocessed, mode=mode)
        return self
    
    def show_dag(self, use_filenames=False):
        """
        Show the DAG for the ETL topology.

        Args:
            use_filenames (bool): Whether or not to use filenames instead of table names.
        """
        return self.__infer_dag(self.topology['etl'], mode=self._mode, use_filenames=use_filenames)
    
    def __handle_compiled_queries(self, config, mode=None):
        if mode is None:
            raise Exception("Error: No mode specified. Please specify a mode.")
        __def_path_dir = ensure_rooting(os.path.join('./scripts', 'compiled', self._name, mode))
        if not os.path.exists(__def_path_dir):
            os.makedirs(__def_path_dir)
        for name in config.keys():
            if mode in config[name] and 'query' in config[name][mode]:
                with open(os.path.join(__def_path_dir, name + '.sql'), 'w') as f:
                    if mode in config[name] and 'query' in config[name][mode]:
                        f.write(config[name][mode]['query'])
    
    def __infer_dag(self, config, mode=None, use_filenames=False):
        queue = []
        # Infer Root Nodes (nodes with no dependencies)
        for node in config.keys():
            if mode in config[node] and 'depends_on' not in config[node][mode]:
                queue.append(node)
            if mode not in config[node]:
                queue.append(node)
        # Add nodes to queue if all dependencies are met
        while set(queue)!=set(config.keys()):
            for node in config.keys():
                if node not in queue:
                    if set(config[node][mode]['depends_on']).issubset(set(queue)):
                        queue.append(node)
        # Identify missing dependency tables
        for node in config.keys():
            if mode in config[node] and 'depends_on' in config[node][mode]:
                missing = [dep for dep in config[node][mode]['depends_on'] if dep not in queue]
                if len(missing)>0:
                    raise Exception(f'Missing dependencies for {node}: {missing}')
        if use_filenames:
            # Use script defined in config, skip if not defined
            requeue = []
            for i, node in enumerate(queue):
                if mode in config[node] and 'script' in config[node][mode]:
                    requeue.append(config[node][mode]['script'])
            queue = requeue
                
            # queue = ['' if 'script' not in config[node][mode] else config[node][mode]['script'] for node in queue]
        return queue

    def __dag_roots(self, config, mode=None):
        roots = []
        for node in config.keys():
            if mode in config[node] and 'depends_on' not in config[node][mode]:
                roots.append(node)
            if mode not in config[node]:
                roots.append(node)
        return roots

    def __wrap_query(self, source, query, method=None):
        """
        Wrap query with a method from the source class.
        
        :param source: A source class that exsavetends the Database class.
        :param query: A query string.
        :param method: A method from the source class. (seed, replace, etc.)
        """
        if method is None:
            raise Exception("Error: No method specified. Please specify a method.")
        # method_patterns is correlated a method and their corresponding query patterns (see sources.py, Redshift class)
        result = source.method_patterns()[method](query) # returns a list of queries
        return ';'.join(result) # returns a single query string
    
    def __prepare_etl(self, config, dag, mode=None, **kwargs):
        def apply_template(query, params): # Jinja Template, this function written to make code easier to read
            return Template(query).render(**params)
        # Execute the DAG
        for node in dag:
            try:
                # Prepare arguments and config
                config_node = config[node][mode]
                this = f"{'public' if 'schema' not in config[node] else config[node]['schema']}.{node}" # relates to {{this}} in query
                args = {**config, 'this':this, **kwargs} # combines config and arguments

                log.info(f"Preparing query for {node}...")
                # Decide to handle as query or script. Script is always a file path. Query is always a SQL string.
                if 'query' in config_node:
                    if 'method' in config_node: # if method is specified, wrap query with method
                        log.info(f"Using method: {config_node['method']}")
                        config_node['query'] = self.__wrap_query(self.source, config_node['query'], method=config_node['method'])
                    config[node][mode]['query'] = apply_template(config_node['query'], args)
                elif 'script' in config_node:
                    with open(ensure_rooting(config_node['script'])) as f:
                        log.info(f"Using script: {config_node['script']}")
                        config_node['query'] = f.read()
                        if 'method' in config_node:
                            log.info(f"Using method: {config_node['method']}")
                            config_node['query'] = self.__wrap_query(self.source, config_node['query'] , method=config_node['method'])
                    config[node][mode]['query'] = apply_template(config_node['query'], args)
            except Exception as e:
                if mode == 'save':
                    config[node].update({mode:{'query':f"SELECT * FROM {node}"}}) # Default to SELECT * FROM table
                    log.error(f'Error preparing query for {node}: {e}')
                    log.warning('Defaulting to preset query.')
                else:
                    log.warning(f'Error preparing query for "{node}" using {mode} mode: SKIPPING ASSUMED ROOT TABLE.')
        return config
    
    def execute(self, tables:list):
        """
        Execute the ETL topology.

        Args:
            tables (list): A list of tables to execute. Order is important.
        """
        # Error Handling - Check if tables are in preprocessed queries
        if self.preprocessed is None:
            raise Exception("Error: No preprocessed queries. Please run prepare() first.")
        if Chasqui.ALL in tables:
            tables = self.dag
        if not isinstance(tables, list):
            raise Exception("Error: tables must be a list.")
        for table in tables:
            if table not in self.preprocessed:
                raise Exception(f"Error: {table} not found in preprocessed queries.")
        
        if self._mode == 'run':
            tables = list(filter(lambda x: x not in self.dag_roots, tables))
            print(tables)

        # Execute Queries
        self.results_ = None
        for table in tables:
            log.info(f"Running {table} using {self.mode_} mode...")
            if self._mode in self.preprocessed[table]:
                steps = self.preprocessed[table][self._mode]['query'].split(';')
                # trim empty strings
                steps = [step for step in steps if step != '']
                for step in steps:
                    # Save mode executes on source, and saves results to a target directory as the specified file type
                    # Results are expected to be pandas dataframes
                    if self._mode == 'save':
                        self.results_ = self._source.execute(step, mode=self._mode)
                        target__ = ensure_rooting(self.topology[self._mode]['target'])
                        file_type__ = None
                        if 'file_type' in self.topology[self._mode]:
                            file_type__ = self.topology[self._mode]['file_type']
                        if not os.path.exists(target__):
                            os.makedirs(target__)
                        __save_types = {
                            'csv': lambda t_: self.results_.to_csv(os.path.join(target__, t_ + '.csv'), index=False),
                            'json': lambda t_: self.results_.to_json(os.path.join(target__, t_ + '.json'), orient='records'),
                            'parquet': lambda t_: self.results_.to_parquet(os.path.join(target__, t_ + '.parquet'), index=False)
                        }
                        if file_type__ is None:
                            __save_types['csv'](table)
                        else:
                            __save_types[file_type__](table)
                    # Run mode only needs to execute, no results should be expected
                    else:
                        self._source.execute(step)
    
    def clean(self, facet):
        """
        Clean the ETL topology.

        Args:
            facet (str): The facet to clean. (compiled, data)
        """
        def remove_directory(dir):
                try:
                    shutil.rmtree(dir)
                except Exception as e:
                    return None
        if '.' in facet:
            facet = ['compiled', 'data']
        if 'compiled' in facet:
            known_target = ensure_rooting(f'./scripts/compiled/{self._name}')
            remove_directory(known_target)
        if 'data' in facet:
            known_targets = list(filter(lambda x: x!=0,[(0 if m not in self.topology else self.topology[m]['target']) for m in ['run', 'save']]))
            for kt in known_targets:
                remove_directory(ensure_rooting(kt))
    
    #====================================================================================================
    # DO NOT TOUCH BELOW ::: REQUIRED FOR CLASS INHERITANCE
    #====================================================================================================
    @property
    def name(self):
        return self._name
    
    @name.setter
    def name(self, name):
        self._name = name

    @property
    def source(self):
        return self._source
    @source.setter
    def source(self, source):
        self._source = source

    @property
    def mode_(self):
        return self._mode
    @mode_.setter
    def mode_(self, mode):
        self._mode = mode

    #====================================================================================================
    # DO NOT TOUCH ABOVE ::: REQUIRED FOR CLASS INHERITANCE
    #====================================================================================================


# A simple setup script
from setuptools import setup, find_packages

setup(name='pycurie',
      version='0.0.23',
      author='Eric DiGioacchino',
      url='https://github.com/EricDiGi/CurieKetl',
      keywords=['etl', 'data', 'data engineering', 'data science', 'mlops', 'devops'],
      classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3"
        ],
        long_description=open('../README.md').read(),
        long_description_content_type='text/markdown',
      packages=find_packages(),
      entry_points={
          'console_scripts': [
              'curie = curie.__main__:main'
          ]
      },
      package_data={
            "curie.defaults": ["*"],
            "curie.defaults.jinja": ["*.md.j2", "*.yaml.j2"]
      }
)


# Author: Eric DiGioacchino
import imp
import logging as log
import re
import os
import subprocess

    
class Database:
    def __init__(self, host, port:int, user, password, database, **kwargs):
        self.host_ = host
        self.port_ = int(port)
        self.user_ = user
        self.password_ = password
        self.database_ = database
        self.kwargs_ = kwargs

    def __repr__(self) -> str:
        return "Database(host={}, port={}, user={}, password={}, database={}, kwargs={})".format(self.host_, self.port_, self.user_, self.password_, self.database_, self.kwargs_)
    

class Redshift(Database):
    def __init__(self, host, port, user, password, database, **kwargs):
        super().__init__(host, port, user, password, database, **kwargs)

        if 'defer_import' in kwargs:
            self.defer_import_ = kwargs['defer_import']
            del self.kwargs_['defer_import']
            try:
                self.defer_import_ = bool(self.defer_import_)
            except:
                log.warn("Warning: defer_import must be a boolean value. Defaulting to False.")
                self.defer_import_ = False

        if self.defer_import_:
            self.__defered_import()
        self.__validate_imports()
        pass

    def __defered_import(self):
        global redshift_connector
        import redshift_connector
    
    def __validate_imports(self):
        # check if psycopg2 and redshift_connector are installed
        try:
            return all([imp.find_module('redshift_connector')])
        except ImportError:
            return False

    def connect(self, help=False):
        # use global variables to connect to redshift
        try:
            self.conn_ = redshift_connector.connect(host=self.host_, port=self.port_, user=self.user_, password=self.password_, database=self.database_, **self.kwargs_)
            log.info("Connected to Redshift...")
            if help:
                print("\n\nExample Usage:\n\nconn = Curie(*args).this_connection.connect()\ncursor = conn.cursor()\ncursor.execute('SELECT * FROM table')\nresults = cursor.fetchall()\n\n")
            return self.conn_
        except Exception as e:
            log.error("Error: Could not connect to Redshift. Check your credentials.")
            log.error(e)
            return None

    def __repr__(self):
        return "Redshift(host={}, port={}, user={}, password={}, database={}, kwargs={})".format(self.host_, self.port_, self.user_, self.password_, self.database_, self.kwargs_)
    
    def method_patterns(self):
        return {
            'seed': lambda q: [q],
            'replace': lambda q: ['DROP TABLE IF EXISTS {{this}}', 'CREATE TABLE {{this}} AS (' + q + ')'],
        }
    def execute(self, query, **kwargs):
        self.results_ = None
        self.conn_ = self.connect()
        self.conn_.autocommit = True
        if self.conn_ is None:
            return None
        cursor = self.conn_.cursor()
        cursor.execute(query)
        if 'mode' in kwargs and kwargs['mode'] == 'save':
            results = cursor.fetch_dataframe()
            self.results_ = results
        cursor.close()
        self.conn_.close()
        return self.results_

class MySQL(Database):
    def __init__(self, host, port, user, password, database, **kwargs):
        super().__init__(host, port, user, password, database, **kwargs)

        if 'defer_import' in kwargs:
            self.defer_import_ = kwargs['defer_import']
            try:
                self.defer_import_ = bool(self.defer_import_)
            except:
                log.warn("Warning: defer_import must be a boolean value. Defaulting to False.")
                self.defer_import_ = False

        if self.defer_import_:
            self.__defered_import()
        self.__validate_imports()
        pass

    def __defered_import(self):
        global mysql
        import mysql
    
    def __validate_imports(self):
        # check if psycopg2 and redshift_connector are installed
        try:
            return all([imp.find_module('redshift_connector')])
        except ImportError:
            return False

    def connect(self):
        # use global variables to connect to redshift
        try:
            del self.kwargs_['defer_import']
            self.conn_ = mysql.connector.connect(host=self.host_, port=self.port_, user=self.user_, password=self.password_, database=self.database_, **self.kwargs_)
            return self.conn_
        except Exception as e:
            log.error("Error: Could not connect to Redshift. Check your credentials.")
            log.error(e)
            return None
    
class MongoDB(Database):
    def __init__(self, **kwargs):
        pass





# // import logging as log
# // import os
# // from typing import Any

# // import yaml

# // import sources
# // from inca import Chasqui
try:
    from .lab import Curie
    from . import mk
except FileNotFoundError as e:
    pass